package org.alltimeflashdreamer.filelister;

import java.io.Serializable;

/**
 * <p>
 *
 * Title: DefineSearchParameters</p> <p>
 *
 * Description: Simple SearchParameters extending class which stores easily accessible
 * parameters for the GUI of the define-search site.</p> <p>
 *
 * Organization: All Time Flash Dreamer</p>
 *
 * @author    Copyright (c) 2002-2004 by <a href="mailto:florian@steinsiepe.ch">Florian
 *      Steinsiepe</a>
 * @version   0.5.1
 */

public class DefineSearchParameters extends SearchParameters implements Serializable
{
    private int iSearchMode = FileListerConfig.NEWSEARCH;

    /**
     * The empty constructor initializes the class with the application's default
     * values.
     *
     * @param iSearchMode  which search mode is active?
     */
    DefineSearchParameters( int iSearchMode )
    {
        super();
        this.iSearchMode = iSearchMode;
    }

    /**
     * This constructor simply copies the values of the given SearchParameters instance,
     * if the given search mode is "search again", else default values are used for
     * the search mode sensitive parameters.
     *
     * @param params       uses this as the settings template
     * @param iSearchMode  which search mode is active?
     */
    DefineSearchParameters( SearchParameters params, int iSearchMode )
    {
        this( iSearchMode );
        if ( params != null )
        {
            setCaseSensitiveSearch( params.isCaseSensitiveSearch() );
            setDirectoryListing( params.isDirectoryListing() );
            setExactMatch( params.isExactMatch() );
            setFileListing( params.isFileListing() );
            setUsingRegExp( params.isUsingRegExp() );
            setSearchString( ( isSearchAgain() ) ? params.getSearchString() : "" );
            setMinimumSize( ( isSearchAgain() ) ? params.getMinimumSize() : NOTSPECIFIED );
            setMinimumMultiplier( ( isSearchAgain() ) ? params.getMinimumMultiplier() : getDefaultMultiplier() );
            setMaximumSize( ( isSearchAgain() ) ? params.getMaximumSize() : NOTSPECIFIED );
            setMaximumMultiplier( ( isSearchAgain() ) ? params.getMaximumMultiplier() : getDefaultMultiplier() );
            setStartingDate( ( isSearchAgain() ) ? params.getStartingDate() : null );
            setEndingDate( ( isSearchAgain() ) ? params.getEndingDate() : null );
        }
    }

    /**
     * Returns whether all types (i.e. files and directory) shall be listed in the
     * result table or not.
     *
     * @return   list all types or not?
     */
    public boolean isListingAllTypes()
    {
        return ( ( isDirectoryListing() && isFileListing() ) || ( !isDirectoryListing() && !isFileListing() ) );
    }

    /**
     * Returns whether <i>only files</i> shall be listed in the results table or
     * not.
     *
     * @return   list files only?
     */
    public boolean isListingFilesOnly()
    {
        return ( !isDirectoryListing() && isFileListing() );
    }

    /**
     * Returns whether <i>only directories</i> shall be listed in the results table
     * or not.
     *
     * @return   list files only?
     */
    public boolean isListingDirectoriesOnly()
    {
        return ( isDirectoryListing() && !isFileListing() );
    }

    /**
     * Returns true only if the search mode is "search within results".
     *
     * @return   current mode is "search within results"?
     */
    public boolean isSearchWithinResults()
    {
        return ( iSearchMode == FileListerConfig.SEARCHWITHINRESULTS );
    }

    /**
     * Returns true only if the search mode is "search again".
     *
     * @return   current mode is "search again"?
     */
    public boolean isSearchAgain()
    {
        return ( iSearchMode == FileListerConfig.SEARCHAGAIN );
    }
}
