package org.alltimeflashdreamer.filelister;

import java.io.File;
import java.io.IOException;
import java.io.Serializable;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Vector;

import org.alltimeflashdreamer.util.ChainedList;
import org.alltimeflashdreamer.util.LimitedList;
import org.alltimeflashdreamer.util.StringUtils;

/**
 * <p>
 *
 * Title: FileLister</p> <p>
 *
 * Description: This class is the core of the FileLister application. It searches
 * the file system downwards, starting with the root directory defined in the FileListerConfig
 * object. The found files (which may be directories too, of course) are collected,
 * if they match the search string's restrictions.</p> <p>
 *
 * Organization: All Time Flash Dreamer</p>
 *
 * @author    Copyright (c) 2002-2004 by <a href="mailto:florian@steinsiepe.ch">Florian
 *      Steinsiepe</a>
 * @version   0.5.1
 */

public class FileLister extends ArrayList
{
    /** Defines the version number of the FileLister (is shown in the JSP output) */
    public final static String VERSIONNUMBER = "0.5";
    /** often used value to find out whether the platform is Windows or not */
    final static boolean ISWINDOWS = ( System.getProperty( "os.name" ).toLowerCase().indexOf( "win" ) >= 0 );
    private final static int MAXRUNNINGDUMPERINSTANCES = 10;
    /** default is index search, i.e. true */
    private boolean bLoadFromDatabase = true;
    /** counts the searched directories (within the recursion) */
    private int iSearchedDirectoryCount = 0;
    /** holds the pointer to the config object */
    private FileListerConfig fileListerConfig;
    /** stores the "last indexed" time of the entry found by getSearchDirectoryID */
    private long lLastIndexed;
    /** stores the total query results count */
    private int iResultCount;
    /** stores the node id of the searched directory */
    private String sNodeID;
    /** stores the total amount of dumped FileListerEntry instances */
    private int iTotalDumpedEntryCount = 0;
    /**
     * registers whether indexing comments shall be printed or not; default is "print
     * comments", i.e. false
     */
    private boolean bQuietMode = false;

    /**
     * If it's the directory input mode, it's checked whether the researching of
     * the given directory is allowed again or not, depending on the index aging
     * parameter of the property file, compared with the "last indexed" value and
     * the current time. If the researching is allowed, it's processed.<br>
     * The database is queried (anyway, even after a rescan of the file system) for
     * the entry with the id of the directory to be searched (which is retrieved
     * first). If this entry isn't defined, the initial archiving process hasn't
     * been done yet which leads to an exception.<br>
     * The found files and/or directories are stored in the ArrayList itself, each
     * represented by a FileListerEntry instance (strictly performance-oriented!).
     *
     * @param fileListerConfig      the previously initialized config object
     * @param bQuietMode            whether to print messages to the System.out or
     *      not (true means "print no messages at all")
     * @throws FileListerException  each possible exception that may occur during
     *      the listing-of-files process
     */
    FileLister( FileListerConfig fileListerConfig, boolean bQuietMode )
        throws FileListerException
    {
        this.fileListerConfig = fileListerConfig;
        this.bQuietMode = bQuietMode;

        //if it's setup or (re)indexing time, do it and break afterwards
        if ( fileListerConfig.isInitialArchivingProcess() )
        {
            runInitialArchivingProcess();
            return;
        }

        //find out the node id of the directory to be searched; this method sets
        //the bLoadFromDatabase-boolean accordingly and finds out whether to read
        //the index or search the file system
        sNodeID = getSearchDirectoryID();

        //rescan file system, if necessary (updates the index too!)
        if ( !bLoadFromDatabase )
            rescanFileSystem();

        //load the entries using query-builder's SQL string
        prepareEntries( 0 );

        //find out the effective result count by requerying the database, if the
        //result count is higher than the (first) LimitedList's entry count
        iResultCount = getEffectiveResultCount();
    }

    /**
     * Returns either the canonical path or, in case of a windows drive letter, the
     * absolute path of the given directory.
     *
     * @param dir           the directory to get the real path of
     * @return              the real path, as described above
     * @throws IOException  thrown on canonical file access
     */
    static String getRealPath( File dir )
        throws IOException
    {
        if ( ISWINDOWS && dir.getParent() == null )
            return StringUtils.replaceBackslashes( dir.getAbsolutePath() );

        return StringUtils.replaceBackslashes( dir.getCanonicalPath() );
    }

    /**
     * Reloads the index from the database, very much like during construction. This
     * method is helpful for using sort capabilities of the SQL server which need
     * a full re-exectution of the underlying query, but with different sort criteria.
     *
     * @throws FileListerException  if some problem occurred during reloading
     */
    void reloadIndex()
        throws FileListerException
    {
        this.clear();
        prepareEntries( fileListerConfig.getStartIndex() );
    }

    /**
     * Returns the indexing time (as long) of the directory node entry of the user's
     * search directory. This value is representative for the whole tree down the
     * search directory, although each directory has its own indexing time.
     *
     * @return   the time to be shown to the client in the results site as the indexing
     *      time
     */
    long getLastIndexed()
    {
        return lLastIndexed;
    }

    /**
     * Returns the total result count of the query that initialized the current instance
     * of FileLister. This value is received by one initial query on the result count
     * (if necessary, i.e. if too many results were found).
     *
     * @return   the result count of the initial query
     */
    public int getResultCount()
    {
        return iResultCount;
    }

    /**
     * Returns the sublist of visible FileListerEntry objects (in the GUI) in <i>
     * of the whole</i> result set received from the underlying query.
     *
     * @return   the FileListerEntry[] containing all the currently visible search
     *      results
     */
    public FileListerEntry[] getVisibleFileListerEntries()
    {
        FileListerEntry[] returnValue = null;
        LimitedList nextList;
        for ( int i = 0; i < this.size(); i++ )
        {
            nextList = ( LimitedList )this.get( i );
            if ( nextList.isInList( fileListerConfig.getStartIndex() ) )
            {
                int iStartIndex = fileListerConfig.getStartIndex() % nextList.getMaxSize();
                int iEndIndex = ( ( iStartIndex + fileListerConfig.getResultsPerSite() ) > nextList.size() ) ? nextList.size() : iStartIndex + fileListerConfig.getResultsPerSite();
                returnValue = ( FileListerEntry[] )nextList.subList( iStartIndex, iEndIndex ).toArray( new FileListerEntry[]{} );
            }
        }
        return ( returnValue == null ) ? new FileListerEntry[]{} : returnValue;
    }

    /**
     * Makes sure that there are enough entries buffered in the memory to allow easy
     * navigation, with only a few reloads.
     *
     * @param iStartIndex           the start index of the position where the user
     *      wants to see entries; this start index is used to load the files
     * @throws FileListerException  thrown if there were problems loading files into
     *      the list
     */
    void prepareEntries( int iStartIndex )
        throws FileListerException
    {
        if ( !isLoaded( iStartIndex ) )
        {
            loadEntriesFromDatabase();
        }
    }

    /**
     * Returns an list containing the selected FileListerEntries within the FileLister
     * or null, if none are selected.
     *
     * @return   an list of selected FileListerEntry objects for zip or null
     */
    ArrayList getFileListerEntriesForZip()
    {
        ArrayList returnValue = new ArrayList();
        LimitedList nextList;
        FileListerEntry nextFLE;
        for ( int i = 0; i < this.size(); i++ )
        {
            nextList = ( LimitedList )this.get( i );
            for ( int j = 0; j < nextList.size(); j++ )
            {
                nextFLE = ( FileListerEntry )nextList.get( j );
                if ( nextFLE.isSelectedForZip() )
                    returnValue.add( nextFLE );
            }
        }
        return returnValue;
    }

    /**
     * If the given FileListerEntry represents a file, 0 is returned, else the index
     * is queried for the child count of the given directory.
     *
     * @param fileListerEntry       the entry to get the child count of
     * @return                      0, if a file, else the database value
     * @throws FileListerException  wrapped SQLException, in case
     */
    int getChildCount( FileListerEntry fileListerEntry )
        throws FileListerException
    {
        if ( fileListerEntry.isFile() )
            return 0;

        int iReturnValue = 0;
        Statement st = getStatement();
        try
        {
            ResultSet rs = st.executeQuery( fileListerConfig.getQueryBuilder().getChildCountString( fileListerEntry.getNodeID() ) );
            if ( rs.next() )
                iReturnValue = rs.getInt( 1 );
            rs.close();
        } catch ( SQLException sqle )
        {
            throw new FileListerException( 15 );
        } finally
        {
            closeStatement( st );
        }

        return iReturnValue;
    }

    /**
     * Returns the total amount of dumped FileListerEntry objects. This number is
     * always reset during creation of a new SimpleList (and of a new FileLister,
     * of course).
     *
     * @return   the total number of dumped entries
     */
    int getTotalDumpedEntryCount()
    {
        return iTotalDumpedEntryCount;
    }

    /**
     * Loops over all lists stored in this class and over all FileListerEntry instances
     * within those lists and selects or unselects all the entries at once, but only
     * if the given string equals "selectall" or "deselectall".
     *
     * @param sZipMode  the string describing whether to select or unselect all the
     *      registered entries
     */
    void processMultiSelectForZip( String sZipMode )
    {
        boolean bSelect = false;
        if ( sZipMode.equals( "selectall" ) )
            bSelect = true;
        else if ( sZipMode.equals( "deselectall" ) )
            bSelect = false;
        else
            return;

        for ( int i = 0; i < this.size(); i++ )
        {
            LimitedList nextList = ( LimitedList )this.get( i );
            for ( int j = 0; j < nextList.size(); j++ )
            {
                ( ( FileListerEntry )nextList.get( j ) ).setSelectedForZip( bSelect );
            }
        }
    }

//-------------------------- private methods ---------------------------------//
    /**
     * Checks whether the index aging parameter allows the searching of the directory
     * again (only in directory search mode!) and return the node_id value for later
     * referencings by finding the database entry having the required seach directory
     * as the directory value; if this entry isn't defined in the database, the initial
     * archiving process hasn't been done yet which leads to the trowing of an exception.
     *
     * @return                      the id of the search directory node
     * @throws FileListerException  wraps a possible SQLException or claims the not
     *      yet done initial archiving process
     */
    private String getSearchDirectoryID()
        throws FileListerException
    {
        bLoadFromDatabase = true;

        //get the search root's node id; use FileListerEntry dummy to transform all
        //the ugly directory stuff (esp. on Windows...)
        String sNodeID = null;
        FileListerEntry fleSearch = new FileListerEntry( fileListerConfig.getStartDirectory(), 0, "" );
        String sSQL = "select * from " + fileListerConfig.getJDBCTableName() + " where " +
                ( ( fleSearch.getDirectory() == null ) ? "isnull( directory )" : "directory = '" + fleSearch.getDirectory() + "'" ) +
                " and name = '" + fleSearch.getName() + "'";
        fleSearch = null;

        //DEBUG: System.out.println( "getSearchDirectoryID: sSQL = " + sSQL );

        Statement st = getStatement();
        try
        {
            ResultSet rs = st.executeQuery( sSQL );
            if ( rs.next() )
            {
                sNodeID = rs.getString( "node_id" );
                lLastIndexed = rs.getLong( "lastindexed" );
                if ( !fileListerConfig.isArchiveMode() )
                {
                    //if the root directory entry is stored, compare the current time with
                    //its last indexed time plus the index aging parameter; if this indicates
                    //that a researching is necessary, update the bLoadFromDatabase flag
                    if ( ( lLastIndexed + 1000L * fileListerConfig.getIndexAging() )
                            < System.currentTimeMillis() )
                        bLoadFromDatabase = false;
                }
            }
            rs.close();
        } catch ( SQLException sqle )
        {
            throw new FileListerException( 15 );
        } finally
        {
            closeStatement( st );
        }

        //throw a "not yet indexed" exception if the given directory wasn't stored in
        //the database/index yet
        if ( sNodeID == null )
            throw new FileListerException( 14 );

        return sNodeID;
    }

    /**
     * Deletes the entries of all child nodes of the node with id sSearchDirectoryNodeID.
     *
     * @param sSeachDirectoryNodeID  the node id of the search directory
     * @throws FileListerException   wraps a SQLException or claims the running of
     *      the initial archiving process
     */
    private void deleteChildren( String sSeachDirectoryNodeID )
        throws FileListerException
    {
        Statement st = getStatement();
        try
        {
            //delete each entry below the search directory from the database, including
            //the search directory entry itself
            int iDeletedRowCount = st.executeUpdate( "delete from " + fileListerConfig.getJDBCTableName() +
                    " where node_id like '" +
                    sSeachDirectoryNodeID + "0%' or node_id = '" + sSeachDirectoryNodeID + "'" );

            //check whether there were deleted entries; if not and it's not the
            //initially indexing, the initial archiving process hasn't been done
            //yet -> throw exception
            if ( !fileListerConfig.isInitialArchivingProcess() && iDeletedRowCount <= 0 )
                throw new FileListerException( 14 );
        } catch ( SQLException sqle )
        {
            closeStatement( st );
            throw new FileListerException( 10, sqle );
        }
        closeStatement( st );
    }

    /**
     * Recursive method which searches the given directory for its contents and creating/storing
     * the corresponding FileListerEntry instances.
     *
     * @param directory             the directory to be searched
     * @param sDirectoryNodeID      the binary tree's node id of the search directory;
     *      this value is updated for each found child node and stored in each newly
     *      created FileListerEntry
     * @param sl                    the list of pointers to add new file lister entries
     *      to
     * @return                      the size of <i>all</i> descendants of the given
     *      directory
     * @throws FileListerException  wraps a possible SecurityException
     */
    private long searchDirectory( File directory, String sDirectoryNodeID, SimpleList sl )
        throws FileListerException
    {
        //count directories and print out command line stuff
        iSearchedDirectoryCount++;
        if ( iSearchedDirectoryCount % 100 == 0 )
        {
            //command line output
            char[] cBackSpaces = new char[new Integer( iSearchedDirectoryCount ).toString().length() + 22];
            Arrays.fill( cBackSpaces, '\b' );
            if ( !bQuietMode )
                System.out.print( new String( cBackSpaces ).toString() + iSearchedDirectoryCount + " directories searched." );
        }

        try
        {
            //binary tree: each found File is a child or the given directory
            //anyway, but each following File is a sibling of the first one
            StringBuffer sbNodeID = new StringBuffer( sDirectoryNodeID );

            //list all contents of each directory and remember the time (for
            //the whole directory because they were found at the same time anyway);
            //on very special occasions, the array of children may be null!
            File[] childFiles = directory.listFiles();
            if ( childFiles == null )
                return 0L;

            //each found File is stored to be found "now"
            long lIndexTime = System.currentTimeMillis();
            long lSizeOfChildren = 0L;

            FileListerEntry nextFLE;
            for ( int i = 0; i < childFiles.length; i++ )
            {
                sbNodeID.append( ( i == 0 ) ? "0" : "1" );
                //create a FileListerEntry only if it's a searchable directory, i.e.
                //not contained in the property file's nogoarea parameter, or if it's
                //not a directory
                nextFLE = new FileListerEntry( childFiles[i], lIndexTime, sbNodeID.toString() );
                if ( isSearchableDirectory( childFiles[i] ) )
                {
                    long lSizeOfChildrenOfThisDirectory = searchDirectory( childFiles[i], sbNodeID.toString(), sl );
                    nextFLE.setSize( lSizeOfChildrenOfThisDirectory );
                    lSizeOfChildren += lSizeOfChildrenOfThisDirectory;
                }
                else
                {
                    lSizeOfChildren += childFiles[i].length();
                    //if it's a directory, it's a forbidden one -> don't add it to the SimpleList
                    if ( childFiles[i].isDirectory() )
                        nextFLE = null;
                }

                if ( nextFLE != null )
                    sl.add( nextFLE );
            }

            return lSizeOfChildren;
        } catch ( SecurityException se )
        {
            throw new FileListerException( 8, se );
        }
    }

    /**
     * Used to compare the given directory to the list of prohibited directories
     * (parameter directory.nogoarea in the properties file). If the given directory
     * is found in the list of no-go areas, it's removed from it to improve the performance
     * of this method.
     *
     * @param dir                   directory to compare with list of prohibited
     *      directories.
     * @return                      Returns true if the given directory is not in
     *      the list.
     * @throws FileListerException  thrown on resolving a canonical path
     */
    private boolean isSearchableDirectory( File dir )
        throws FileListerException
    {
        if ( !dir.isDirectory() || isSymbolicLink( dir ) )
            return false;

        try
        {
            String sRealPath = getRealPath( dir );
            if ( fileListerConfig.getNoGoAreas().contains( sRealPath ) )
            {
                fileListerConfig.getNoGoAreas().remove( sRealPath );
                return false;
            }
        } catch ( IOException ioe )
        {
            //DEBUG: ioe.printStackTrace();
            return false;
        }

        return true;
    }

    /**
     * Gets a connection from the pool and returns a Statement created by this connection
     *
     * @return                      the newly created Statement
     * @throws FileListerException  wrapped SQLException, in case
     */
    private Statement getStatement()
        throws FileListerException
    {
        try
        {
            //create a JDBC connection and returns the newly created Statement
            Connection con = PoolHandler.getPool().getConnection();
            return con.createStatement();
        } catch ( SQLException sqle )
        {
            throw new FileListerException( 21, sqle );
        }
    }

    /**
     * Gets a connection from the pool and returns a PreparedStatement created by
     * this connection, initialized with the given SQL string.
     *
     * @param sSQL                  the SQL string to prepare
     * @return                      the newly created PreparedStatement
     * @throws FileListerException  wrapped SQLException, in case
     */
    private PreparedStatement prepareStatement( String sSQL )
        throws FileListerException
    {
        try
        {
            //create a JDBC connection and returns the newly prepared statement
            Connection con = PoolHandler.getPool().getConnection();
            return con.prepareStatement( sSQL );
        } catch ( SQLException sqle )
        {
            throw new FileListerException( 21, sqle );
        }
    }

    /**
     * Closes the given Statement and returns its connection to the pool. Each possibly
     * upcoming exception is ignored.
     *
     * @param st  the (Prepared)Statement to close
     */
    private void closeStatement( Statement st )
    {
        try
        {
            Connection con = st.getConnection();
            st.close();
            PoolHandler.getPool().returnConnection( con );
        } catch ( Exception ex )
        {
        }
    }

    /**
     * Finds out the root directory first. If no virtual root is defined in the property
     * file and the underlying machine runs Windows, the whole list of mapped drives
     * is searched, else the absolute or the defined virtual root directory is indexed.
     * <br>
     * Each found "entry" (which may be a file, a directory or a symbolic link) is
     * stored in one table row. Symbolic links are ignored to avoid duplicates in
     * the index.
     *
     * @throws FileListerException  wraps each possibly upcoming exception
     */
    private void runInitialArchivingProcess()
        throws FileListerException
    {
        File[] rootDirectories = null;

        //create new temp table, is later renamed to fileListerConfig.getJDBCTableName()
        String sTempTableName = fileListerConfig.getJDBCTableName() + "_" + System.currentTimeMillis();
        Statement st = getStatement();
        try
        {
            st.executeUpdate( fileListerConfig.getQueryBuilder().getCreateTableString( sTempTableName ) );
            closeStatement( st );
        } catch ( SQLException sqle )
        {
            closeStatement( st );
            throw new FileListerException( 15, sqle );
        }

        //if no virtual root is defined and we're running on windows,
        //index the whole list of mapped drives, else index the virtual root
        if ( StringUtils.isEmpty( fileListerConfig.getVirtualRoot() ) )
        {
            if ( ISWINDOWS )
            {
                rootDirectories = File.listRoots();
                for ( int i = 0; i < rootDirectories.length; i++ )
                {
                    if ( fileListerConfig.getNoGoAreas().contains( rootDirectories[i].getAbsolutePath() ) )
                        rootDirectories[i] = null;
                }
            }
            else
                rootDirectories = new File[]{new File( "/" )};
        }
        else
            rootDirectories = new File[]{new File( fileListerConfig.getVirtualRoot() )};

        //start the recursive search(es) which add(s) each found and accepted entry
        //to the array list (i.e. this class itself)
        StringBuffer sbRootIndexes = new StringBuffer();
        SimpleList sl = null;
        for ( int i = 0; i < rootDirectories.length; i++ )
        {
            //even virtual root directories may be listed in "directory.nogoarea"...
            if ( rootDirectories[i] == null || !isSearchableDirectory( rootDirectories[i] ) )
                continue;

            sbRootIndexes.append( ( i == 0 ) ? "0" : "1" );
            FileListerEntry topFLE = new FileListerEntry( rootDirectories[i], System.currentTimeMillis(),
                    sbRootIndexes.toString() );

            //initialize a SimpleList, keep the top/root directory pointer until the very end of the
            //indexing process - this is needed to calculate the total/overall directory size(s) -
            //and finally dump the whole list, including the top entry
            sl = new SimpleList( sTempTableName );
            topFLE.setSize( searchDirectory( rootDirectories[i], sbRootIndexes.toString(), sl ) );
            sl.add( topFLE );
            sl.dump( true );
        }

        //clear "hot" table and move temp table data to original table
        try
        {
            st = getStatement();
            ResultSet rs = st.executeQuery( fileListerConfig.getQueryBuilder().getListAllTablesString( fileListerConfig.getJDBCTableName() ) );
            boolean bDropTable = rs.next();

            //set auto-commit to false for a while, to make sure the dropping and moving
            //is executed "at once" within the SQL server (thread safety reasons)
            st.getConnection().setAutoCommit( false );
            if ( bDropTable )
                st.executeUpdate( fileListerConfig.getQueryBuilder().getDropTableString( fileListerConfig.getJDBCTableName() ) );
            st.executeUpdate( fileListerConfig.getQueryBuilder().getRenameTableString( sTempTableName, fileListerConfig.getJDBCTableName() ) );
            st.getConnection().commit();
            st.getConnection().setAutoCommit( true );
            closeStatement( st );
        } catch ( SQLException sqle )
        {
            closeStatement( st );
            throw new FileListerException( 28, sqle );
        }
    }

    /**
     * Returns whether the file's absolute and canonical paths are equal; if not,
     * file represents a symbolic link.
     *
     * @param file                  File object to check
     * @return                      whether it's a symbolic link or not
     * @throws FileListerException  thrown on resolving a canonical path
     */
    private boolean isSymbolicLink( File file )
        throws FileListerException
    {
        //only non-windows-machines may use symbolic links!
        if ( ISWINDOWS )
            return false;

        try
        {
            return !( file.getCanonicalPath().equals( file.getAbsolutePath() ) );
        } catch ( IOException ioe )
        {
            throw new FileListerException( 25, ioe );
        }
    }

    /**
     * This method is called <i>only if</i> the URL parameter parsing process found
     * out that a rescanning of the file system (partially, i.e. starting with the
     * directory referenced by sNodeID) is required.
     *
     * @throws FileListerException  wrapper for each possibly upcoming exception
     */
    private void rescanFileSystem()
        throws FileListerException
    {
        //start the recursive search which adds each found and accepted entry
        //to the SimpleList (which starts dumping on its own, if necessary)
        File startDirectory = fileListerConfig.getStartDirectory();
        deleteChildren( sNodeID );
        long lReindexingTime = System.currentTimeMillis();
        FileListerEntry topFLE = new FileListerEntry( startDirectory, lReindexingTime, sNodeID );

        //initialize a SimpleList, keep the top/root directory pointer until the very end of the
        //indexing process - this is needed to calculate the total/overall directory size(s) -
        //and finally dump the whole list, including the top entry
        SimpleList sl = new SimpleList( fileListerConfig.getJDBCTableName() );
        topFLE.setSize( searchDirectory( startDirectory, sNodeID, sl ) );
        sl.add( topFLE );
        sl.dump( true );

        //set last indexed time after everything is done
        lLastIndexed = lReindexingTime;
    }

    /**
     * Asks the query builder for the database-specific SQL string, executes it and
     * stores the results in the internal list, each represented by an instance of
     * FileListerEntry and having an index entry in the Hashtable which stores the
     * index/position pairs.
     *
     * @throws FileListerException  wraps possible SQLExceptions
     */
    private void loadEntriesFromDatabase()
        throws FileListerException
    {
        //remove older entries if necessary
        checkEntryCount();

        //initialize a LimitedList and fill it with the results
        LimitedList list = new LimitedList( fileListerConfig.getBufferSize(),
                fileListerConfig.getStartIndex() - ( fileListerConfig.getStartIndex() % fileListerConfig.getBufferSize() ) );
        setEntryLimitations( list.getStartIndex(), list.getStartIndex() + list.getMaxSize() );

        //get the SQL string from the query builder
        String sSQL = fileListerConfig.getQueryBuilder().getUserSearchString( sNodeID, false );

        try
        {
            //read the data from the index and store data objects for each table row
            Statement st = getStatement();
            ResultSet rs = st.executeQuery( sSQL );
            while ( rs.next() && !list.isFull() )
            {
                //don't add the search directory itself to the list of entries!
                String sNextNodeID = rs.getString( "node_id" );
                if ( !sNextNodeID.equals( sNodeID ) )
                {
                    list.add( new FileListerEntry( rs.getString( "name" ), rs.getString( "directory" ),
                            rs.getLong( "size" ), rs.getLong( "lastmodified" ), rs.getString( "type" ),
                            rs.getLong( "lastindexed" ), sNextNodeID ) );
                }
            }
            rs.close();
            closeStatement( st );
        } catch ( SQLException sqle )
        {
            throw new FileListerException( 10, sqle );
        }

        //add the new LimitedList to the internal list
        this.add( list );
    }

    /**
     * Simply sets the two limit values to the query builder.
     *
     * @param iLowerLimit  lower limit to set
     * @param iUpperLimit  upper limit to set
     */
    private void setEntryLimitations( int iLowerLimit, int iUpperLimit )
    {
        fileListerConfig.getQueryBuilder().setLowerLimit( iLowerLimit );
        fileListerConfig.getQueryBuilder().setUpperLimit( iUpperLimit );
    }

    /**
     * Returns whether the FileListerEntry having the given index is (pre)loaded
     * or not
     *
     * @param iIndex  the index of the entry
     * @return        it's already loaded?
     */
    private boolean isLoaded( int iIndex )
    {
        LimitedList nextList;
        for ( int i = 0; i < this.size(); i++ )
        {
            nextList = ( LimitedList )this.get( i );
            if ( nextList.isInList( iIndex ) )
                return true;
        }
        return false;
    }

    /**
     * If the currently loaded/cached entry is equal or higher than the maximum allowed
     * value (from the property file), the oldest LinkedList is removed from the
     * internal list.
     */
    private void checkEntryCount()
    {
        if ( getCachedEntriesCount() >= fileListerConfig.getMaxFileListerObjects() )
            this.remove( 0 );
    }

    /**
     * Counts all the cached entries by summarizing their entry counts.
     *
     * @return   the amount of all stored/cached entries
     */
    private int getCachedEntriesCount()
    {
        LimitedList nextList;
        int iReturnValue = 0;
        for ( int i = 0; i < this.size(); i++ )
            iReturnValue += ( ( LimitedList )this.get( i ) ).size();
        return iReturnValue;
    }

    /**
     * Requeries the database only if the first LimitedList is full.<br>
     * This method must be called during constructing of the class only, out of performance
     * reasons, especially for queries with large result sets!
     *
     * @return                      the first LimitedList's entry count or some db
     *      value
     * @throws FileListerException  if the database query failed somehow
     */
    private int getEffectiveResultCount()
        throws FileListerException
    {
        LimitedList firstList = ( LimitedList )this.get( 0 );
        if ( !firstList.isFull() )
            return firstList.size();

        int iReturnValue = 0;
        Statement st = getStatement();
        try
        {
            ResultSet rs = st.executeQuery( fileListerConfig.getQueryBuilder().getUserSearchString( sNodeID, true ) );
            if ( rs.next() )
                iReturnValue = rs.getInt( 1 );
            rs.close();
            closeStatement( st );
        } catch ( SQLException sqle )
        {
            closeStatement( st );
            throw new FileListerException( 15, sqle );
        }
        return iReturnValue;
    }

//-------------------- inner class Dumper ------------------------//
    /**
     * <p>
     *
     * Title: Dumper</p> <p>
     *
     * Description: This thread dumps the list of FileListerEntries that were initialized
     * during scanning of the file system, parallely to the scanning itself. The
     * idea of this mechanism is to avoid too much memory load.</p> <p>
     *
     * Organization: All Time Flash Dreamer</p>
     *
     * @author    Copyright (c) 2002 by <a href="mailto:florian@steinsiepe.ch">Florian
     *      Steinsiepe</a>
     * @version   0.5
     */
    private class Dumper extends Thread implements Serializable
    {
        private PreparedStatement pst;
        private SimpleList sl;
        private boolean bDumpAll;
        private int iID;

        /**
         * Constructor prepares the insert statement and stores some pointers.
         *
         * @param sl                    the pointer of the list to dump
         * @param bDumpAll              whether dump the list including the very
         *      last entry or not
         * @throws FileListerException  each possible exception that may occur during
         *      the dumping-of-FileListerEntries process
         */
        private Dumper( SimpleList sl, boolean bDumpAll )
            throws FileListerException
        {
            pst = prepareStatement( fileListerConfig.getQueryBuilder().getInsertString( sl.sTableName ) );

            this.sl = sl;
            this.bDumpAll = bDumpAll;
            this.iID = sl.dumpers.size() + 1;
        }

        /**
         * The run methods asks the SimpleList for the next element, as long as there
         * are more of them. Each entry retrieved is inserted into the database,
         * using the precompiled PreparedStatement instance.<br>
         * Because this is a (standalone) thread and the run method isn't allowed
         * to throw exceptions, the only possibility is to write out the stack trace
         * and stop inserting which leads to a unstable state of the index/application.
         */
        public void run()
        {
            int iDumpedCtr = 0;
            FileListerEntry entry;
            if ( !bQuietMode )
                System.out.println( "\nDumper #" + iID + " starting... (free memory: " +
                        ( Runtime.getRuntime().freeMemory() / 1024L ) + " KB, " +
                        sl.getEntryCount() + " FileListerEntry objects)" );

            try
            {
                try
                {
                    //get next entry and tell whether to include the
                    //last entry of the list or not -> getNextElement
                    //may return null -> break
                    while ( ( entry = sl.getNextElement( bDumpAll ) ) != null )
                    {
                        //store next entry
                        pst.setString( 1, entry.getName() );
                        pst.setString( 2, entry.getDirectory() );
                        pst.setLong( 3, entry.getSize() );
                        pst.setLong( 4, entry.getLastModified() );
                        pst.setString( 5, entry.getType() );
                        pst.setLong( 6, entry.getLastIndexed() );
                        pst.setString( 7, entry.getNodeID() );
                        if ( !( pst.executeUpdate() > 0 ) )
                            throw new FileListerException( 20 );

                        iDumpedCtr++;
                    }
                    closeStatement( pst );
                } catch ( SQLException sqle )
                {
                    throw new FileListerException( 20, sqle );
                }

                if ( !bQuietMode )
                    System.out.println( "Dumper #" + iID + " finished: " + iDumpedCtr + " entries dumped." );
            } catch ( FileListerException fe )
            {
                if ( !bQuietMode )
                {
                    System.out.println( "" );
                    fe.printStackTrace();
                    if ( fe.getRootCause() != null )
                        fe.getRootCause().printStackTrace();
                }
            } catch ( Exception ex )
            {
                if ( !bQuietMode )
                {
                    System.out.println( "\nAn unexpected error occurred!" );
                    ex.printStackTrace();
                }
            }

            //remove Dumper pointer from SimpleList's Dumper pointer list
            //and add the amount of dumped entries to the total number
            sl.dumpers.remove( this );
            iTotalDumpedEntryCount += iDumpedCtr;
        }
    }

//-------------------- inner class SimpleList ------------------------//
    /**
     * <p>
     *
     * Title: SimpleList</p> <p>
     *
     * Description: The SimpleList is called simple because it's just a pointer chain,
     * starting with the first pointer received during construction. Each entry maps
     * itself to the next pointer/object. The access to the entry count parameter
     * as well as to the elements of the list is synchronized, to allow the parallel
     * adding of new elements as well as the dumping of old ones.</p> <p>
     *
     * Organization: All Time Flash Dreamer</p> <p>
     *
     * </p>
     *
     * @author    Copyright (c) 2002 by <a href="mailto:florian@steinsiepe.ch">Florian
     *      Steinsiepe</a>
     * @version   0.5
     */
    private class SimpleList extends ChainedList implements Serializable
    {
        private int iStartDumpThreshold;
        private Vector dumpers = new Vector( MAXRUNNINGDUMPERINSTANCES );
        private String sTableName;

        /**
         * Constructor receives the table name to dump the entries to. To start the
         * pointer chain, a dummy FileListerEntry is initialized and set as the first
         * element. This dummy may never occur in within the results of a search
         * because it's not a child of the virtual root entry.
         *
         * @param sTableName  the table name to dump those entries to
         */
        private SimpleList( String sTableName )
        {
            super( new FileListerEntry( "dummy", "dummy", 0L, 0L, "D", 0L, "dummy" ) );
            iStartDumpThreshold = fileListerConfig.getMaxFileListerObjects() / 2;
            this.sTableName = sTableName;
            iTotalDumpedEntryCount = 0;
        }

        /**
         * Override of the superclass' add method, to check whether the entry count
         * becomes too high. If so, a new Dumper is started which flushes the list's
         * contents to the database.
         *
         * @param entry                 the new last element of the list
         * @throws FileListerException  thrown on initializing a entry-flushing Dumper
         *      instance on connecting the database
         */
        private void add( FileListerEntry entry )
            throws FileListerException
        {
            super.add( entry );

            //flushing to database (free memory!)
            if ( isNeededToStartDumper() )
                this.dump( false );
        }

        /**
         * Calculates from time to time whether it's necessary to start a new Dumper
         * instance to reduce the entry count of the list / the memory consume.
         *
         * @return   too many objects in the list?
         */
        private boolean isNeededToStartDumper()
        {
            if ( getEntryCount() >= iStartDumpThreshold && ( getEntryCount() % ( iStartDumpThreshold / MAXRUNNINGDUMPERINSTANCES ) == 0 ) )
            {
                int iLoopCtr = dumpers.size();
                int iCompareValue = iStartDumpThreshold;
                while ( iLoopCtr > 0 )
                {
                    iCompareValue += iStartDumpThreshold / ( dumpers.size() + 1 );
                    iLoopCtr--;
                }
                return ( getEntryCount() >= iCompareValue );
            }

            return false;
        }

        /**
         * Calls the superclass' remove method which removes the first element of
         * the chain and returns it. The boolean indicates whether the last element
         * is allowed to be returned too or not.
         *
         * @param bIncludingLast  include the last entry?
         * @return                the next entry in the list or null, if the list
         *      is empty (or empty - 1)
         */
        private FileListerEntry getNextElement( boolean bIncludingLast )
        {
            //avoid returning the last entry, if prohibited by bIncludingLast
            if ( this.isEmpty() || ( !bIncludingLast && getEntryCount() <= 1 ) )
                return null;

            //return the removed element
            return ( FileListerEntry )this.remove();
        }

        /**
         * Starts the next dumper and lets it work. Only if the given boolean says
         * "finish", the method waits until this is done before returning.
         *
         * @param bDumpAll              whether to dump the whole list or not
         * @throws FileListerException  database problems?
         */
        private void dump( boolean bDumpAll )
            throws FileListerException
        {
            //avoid initializing too many dumpers
            if ( dumpers.size() >= MAXRUNNINGDUMPERINSTANCES )
                return;

            //...then add the new Dumper
            Dumper dumper = new Dumper( this, bDumpAll );
            dumpers.add( dumper );
            dumper.start();

            //if "dump until finished", wait until the dumper has done its job!
            if ( bDumpAll )
            {
                while ( this.hasMoreElements() )
                {
                    try
                    {
                        Thread.sleep( 500 );
                    } catch ( InterruptedException ie )
                    {
                        //Bad dreams?
                    }
                }
            }
        }
    }
}
