package org.alltimeflashdreamer.filelister;

import java.util.ArrayList;

import org.alltimeflashdreamer.filelister.xml.FileListerXMLHandler;

/**
 * <p>
 *
 * Title: InitialIndexer</p> <p>
 *
 * Description: Runs and controls the general, so called "initial" archiving process
 * of the FileLister application. This process scans the file system and builds up
 * the index in the database which then may be searched through the web module.</p>
 * <p>
 *
 * Organization: All Time Flash Dreamer</p>
 *
 * @author    Copyright (c) 2002-2004 by <a href="mailto:florian@steinsiepe.ch">Florian
 *      Steinsiepe</a>
 * @version   0.5.1
 */

public class InitialIndexer
{
    private final static String CONFIGFILE = FileListerXMLHandler.CONFIGFILE;

    /**
     * The constructor may be called through the command line and calls the FileLister
     * through the "non-servlet" constructor which automatically runs the initial
     * indexing process.
     *
     * @param sPropertyFileDir      the directory where the property file is stored
     * @param bQuietMode            whether to print messages to the System.out or
     *      not (true means "print no messages at all")
     * @throws FileListerException  wrapper for a config file error or a SQL server
     *      connection exception
     */
    public InitialIndexer( String sPropertyFileDir, boolean bQuietMode )
        throws FileListerException
    {
        //build the PropertyFileParameters and the PoolHandler's ConnectionPool
        PropertyFileParameters params = null;
        try
        {
            params = new PropertyFileParameters( sPropertyFileDir );
            PoolHandler.initializePool( params );
        } catch ( Exception ex )
        {
            if ( ex instanceof FileListerException )
                throw ( FileListerException )ex;

            throw new FileListerException( 22, ex );
        }

        //initialize the FileListerConfig through its non-servlet constructor
        FileListerConfig fileListerConfig = new FileListerConfig( params );
        if ( !bQuietMode )
            System.out.println( "\nSuccessfully initialized the configuration and the database connection pool." );

        //run the initial archiving/indexing process by simply initalizing a FileLister instance
        if ( !bQuietMode )
            System.out.println( "Starting the indexing process..." );
        FileLister fl = new FileLister( fileListerConfig, bQuietMode );

        if ( !bQuietMode )
            System.out.println( "Successfully finished the indexing process (totally " + fl.getTotalDumpedEntryCount() +
                    " entries dumped).\nFileLister is now ready to be accessed through the web browser / servlet runner.\n\n" );
    }

    /**
     * Main initializes the "normal" constructor and handles the various possible
     * exceptions of the FileHandler application as well as the unexpected, but nevertheless
     * unloved OutOfMemoryError. This error may come up in case the "system.maxfilelisterobjects"
     * parameter of the property file is too high.<br>
     * The list command line arguments must contain the path to the config file and
     * may contain the "quiet mode" parameter. The order of the arguments is optional.
     *
     * @param args  must contain the path to the config file and may contain the
     *      quiet mode switch
     */
    public static void main( String[] args )
    {
        //check length of args
        if ( args == null || args.length == 0 || args.length > 2 )
        {
            printUsage();
            return;
        }

        //add all the command line arguments to a list
        ArrayList arrayCommandLineArguments = new ArrayList();
        for ( int i = 0; i < args.length; i++ )
            arrayCommandLineArguments.add( args[i] );

        //find out whether it's the quiet mode or not
        boolean bQuietMode = false;
        if ( arrayCommandLineArguments.size() > 1 )
        {
            for ( int i = 0; i < arrayCommandLineArguments.size(); i++ )
            {
                if ( isIndicatingQuietMode( ( String )arrayCommandLineArguments.get( i ) ) )
                {
                    bQuietMode = true;
                    arrayCommandLineArguments.remove( i );
                    break;
                }
            }
            if ( !bQuietMode )
            {
                printUsage();
                return;
            }
        }

        //start the initial indexing process
        try
        {
            new InitialIndexer( ( String )arrayCommandLineArguments.get( 0 ), bQuietMode );
        } catch ( Exception ex )
        {
            ex.printStackTrace();
            if ( ex instanceof FileListerException )
            {
                FileListerException fle = ( FileListerException )ex;
                if ( fle.getRootCause() != null )
                {
                    System.out.println( "Root cause:" );
                    fle.getRootCause().printStackTrace();
                }
            }
        } catch ( OutOfMemoryError oome )
        {
            System.out.println( "\n\nOut of memory: Please reduce the \"system.maxfilelisterobjects\" " +
                    "parameter in the property file named " + CONFIGFILE + "!\nSorry." );
        }
    }

    /** Prints the usage text. */
    private static void printUsage()
    {
        System.out.println( "Usage: InitialIndexer [OPTION]" );
        System.out.println( "Run the FileLister's indexing process.\n" );
        System.out.println( "  -q, --quiet, /q    Do not print any output\n" );
        System.out.println( "You may also call " );
        System.out.println( "java org.alltimeflashdreamer.filelister.InitialIndexer [OPTION] PATH" );
        System.out.println( "where \"PATH\" defines the directory where the " + CONFIGFILE + " file is stored." );
        System.out.println( "Note: This " + CONFIGFILE + " must be accessible to the InitialIndexer!" );
    }

    /**
     * Returns whether the given command line argument indicates the quiet mode or
     * not.
     *
     * @param sCommandLineArgument  the command line argument to check
     * @return                      indicating quiet mode?
     */
    private static boolean isIndicatingQuietMode( String sCommandLineArgument )
    {
        return ( sCommandLineArgument.equals( "-q" ) ||
                sCommandLineArgument.equals( "--quiet" ) ||
                sCommandLineArgument.equalsIgnoreCase( "/q" ) );
    }
}
