package org.alltimeflashdreamer.filelister;

import org.alltimeflashdreamer.util.StringUtils;

/**
 * <p>
 *
 * Title: JSPMethods</p> <p>
 *
 * Description: This class collects all the methods that are used in the JSP sites
 * in a statically accessible manner.</p> <p>
 *
 * Organization: All Time Flash Dreamer</p> <p>
 *
 * </p>
 *
 * @author    Copyright (c) 2002-2004 by <a href="mailto:florian@steinsiepe.ch">Florian
 *      Steinsiepe</a>
 * @version   0.5.1
 */

public class JSPMethods
{
    /**
     * Used by definesearch.jsp: Returns the appropriate URL parameter for the next
     * search mode, i.e. "simple" or "advanced", depending on the current search
     * mode.
     *
     * @param fileListerConfig  the magic config object
     * @return                  the URL parameter describing the next search mode
     */
    public static String getModeDescriptor( FileListerConfig fileListerConfig )
    {
        return ( ( fileListerConfig.isAdvancedSearch() ) ? "simple" : "advanced" );
    }

    /**
     * Used by viewresults.jsp: Returns the site title, for both a certain result
     * count and no results.
     *
     * @param fileListerConfig  the magic config object
     * @return                  the fully formatted site title.
     */
    public static String getSiteTitle( FileListerConfig fileListerConfig )
    {
        return ( ( fileListerConfig.getFileLister().getResultCount() > 0 ) ? "View Results" : "No Results" );
    }

    /**
     * Used by viewresults.jsp: Returns the action attribute's value for the form
     * tag.
     *
     * @param fileListerConfig  the magic config object
     * @return                  either a link to the download or the results site
     *      URL
     */
    public static String getActionTag( FileListerConfig fileListerConfig )
    {
        return ( fileListerConfig.isAllowedToZip() )
                ? fileListerConfig.getDownloadURL()
                : fileListerConfig.getRequestURL( false );
    }

    /**
     * Used by viewresults.jsp: Calls SearchParameters.formatSize <i>except</i> it's
     * a directory and its size is 0.
     *
     * @param lSizeInBytes  the size of the entry in bytes
     * @param bIsDirectory  whether its a directory or not
     * @return              the formatted size string
     * @see                 StringUtils#formatSize
     */
    public static String formatSize( long lSizeInBytes, boolean bIsDirectory )
    {
        if ( bIsDirectory && lSizeInBytes == 0L )
            return "";

        return StringUtils.formatSize( lSizeInBytes );
    }

    /**
     * Used by viewresults.jsp: Builds the link for each entry in the "Name" column
     * (direct access to this file or directory).
     *
     * @param fle                   the entry to get the link of
     * @param fileListerConfig      the magic config object
     * @return                      the fully formatted link target string
     * @throws FileListerException  missing configurations?
     */
    public static String getNameColumnLink( FileListerEntry fle, FileListerConfig fileListerConfig )
        throws FileListerException
    {
        if ( StringUtils.isEmpty( fileListerConfig.getLinkPrefix() ) )
            return "";
        return new StringBuffer( fileListerConfig.getLinkPrefix() ).append(
                fileListerConfig.getDirectoryOutput( fle.getDirectory() ) ).append( fle.getName() ).toString();
    }

    /**
     * Used by viewresults.jsp: Builds the link for each entry in the "Location"
     * column (direct access to this directory). This link is forced to end with
     * a "/" (forward slash).
     *
     * @param fle                   the entry to get the link of
     * @param fileListerConfig      the magic config object
     * @return                      the fully formatted link target string
     * @throws FileListerException  missing configurations?
     */
    public static String getDirectoryColumnLink( FileListerEntry fle, FileListerConfig fileListerConfig )
        throws FileListerException
    {
        if ( StringUtils.isEmpty( fileListerConfig.getLinkPrefix() ) )
            return "";
        return new StringBuffer( fileListerConfig.getLinkPrefix() ).append(
                fileListerConfig.getDirectoryOutput( fle.getDirectory() ) ).toString();
    }

    /**
     * Used by viewresults.jsp: Returns the fully configured navigation link for
     * a table header column's <i>name</i> .
     *
     * @param fileListerConfig  the magic config object
     * @param sColumnName       the name of the column to configure
     * @return                  the fully configured navigation link
     */
    public static String getColumnLink( FileListerConfig fileListerConfig, String sColumnName )
    {
        return getNavigationLink( fileListerConfig, fileListerConfig.getNextSortOrder( sColumnName ) );
    }

    /**
     * Used by viewresults.jsp: Helper method for getColumnLink, builds one link
     * target (of three).
     *
     * @param fileListerConfig  the magic config object
     * @param sNextSortOrder    how to sort after the click on this link?
     * @return                  the fully configured link target
     */
    public static String getNavigationLink( FileListerConfig fileListerConfig, String sNextSortOrder )
    {
        return new StringBuffer( fileListerConfig.getRequestURL( false ) ).append( "&column=" )
                .append( sNextSortOrder ).toString();
    }

    /**
     * Used by different JSP's: Builds one image src attribute value based on the
     * given name.
     *
     * @param sName  how to name the link
     * @return       the full image src value
     */
    public static String getImageSrc( String sName )
    {
        return new StringBuffer( "images/" ).append( sName ).append( ".gif" ).toString();
    }

//---------------------------------- private methods ------------------------------//
    /** Prevents this class from being instantiated. */
    private JSPMethods() { }
}
