package org.alltimeflashdreamer.filelister;

import java.io.Serializable;
import java.util.ArrayList;

/**
 * <p>
 *
 * Title: MySQLQueryBuilder</p> <p>
 *
 * Description: Implementation of the SQLQueryBuilder class, using proprietary SQL
 * of the MySQL 3.23 series. For later version, nothing should change, but there's
 * no guarantee at all...</p> <p>
 *
 * Organization: All Time Flash Dreamer</p>
 *
 * @author    Copyright (c) 2002-2004 by <a href="mailto:florian@steinsiepe.ch">Florian
 *      Steinsiepe</a>
 * @version   0.5.1
 */

public class MySQLQueryBuilder extends SQLQueryBuilder implements Serializable
{
    /**
     * Empty constructor, as demanded by the superclass specification.
     *
     * @see   org.alltimeflashdreamer.filelister.SQLQueryBuilder
     */
    public MySQLQueryBuilder() { }

//-------------------------- implementations of abstract methods -------------------------------//
    /**
     * Queries the FileListerConfig for the information needed to build a very specific
     * (and therefore Java-unburdening) SQL string.<br>
     * An example SQL string would be:<br>
     * <code>&nbsp;select * from content<br>
     * &nbsp;where node_id like binary '0011%' and name like '%.html'<br>
     * &nbsp;order by name asc limit 0, 500</code>
     *
     * @param sNodeID          the id of the directory to be searched
     * @param bForResultCount  whether the SQL string leads to the returning of the
     *      result count only or to the returning of the full table data
     * @return                 the SQL string which reduces the amount of returned
     *      entries as much as possible
     * @see                    org.alltimeflashdreamer.filelister.SQLQueryBuilder#getUserSearchString
     */
    public String getUserSearchString( String sNodeID, boolean bForResultCount )
    {
        ArrayList arraySearchParams = getSearchParameters();

        //build start of query string defining who's child nodes shall be searched
        //(node id and type are overall settings and are set first in the list of "where"
        //conditions)
        String sStart = ( !bForResultCount ) ? "select * from " : "select count(id) from ";
        StringBuffer sb = new StringBuffer( sStart ).append( getFileListerConfig().getJDBCTableName() );
        addNodeIDFilter( sb, sNodeID );
        addTypeFilter( sb );

        //add, SearchParameters instance after instance, the search parameters/restrictions
        SearchParameters nextParams;
        for ( int i = 0; i < arraySearchParams.size(); i++ )
        {
            nextParams = ( SearchParameters )arraySearchParams.get( i );

            //add the fully configured search string, including keywords (binary, regexp)
            addSearchString( sb, nextParams );

            //add date and size conditions, if any
            addSizeConditions( sb, nextParams );
            addDateConditions( sb, nextParams );
        }

        //append order column and order direction to the query string
        addSorting( sb );

        //value checks and adding of result limitations, if defined
        sb.append( " " ).append( getLimitationString() );

        //DEBUG: System.out.println( "MySQLBuilder user search query string:\n\"" + sb.toString() + "\"" );

        //return whole SQL string
        return sb.toString();
    }


    /**
     * Builds and returns the MySQL query string used to build an indexing table
     * within the database where the index tables are stored. Because a new table
     * is created for each separate index, the name of the new table is passed as
     * a method attribute.<br>
     * Such a table consists of the following columns:
     * <ul>
     *   <li> id (primary key)
     *   <li> name (full text index)
     *   <li> directory
     *   <li> size
     *   <li> lastmodified
     *   <li> type
     *   <li> lastindexed
     *   <li> node_id
     * </ul>
     *
     *
     * @param sTableName  the name of the index table to be created
     * @return            the fully qualified MySQL string
     * @see               org.alltimeflashdreamer.filelister.SQLQueryBuilder#getCreateTableString
     */
    public String getCreateTableString( String sTableName )
    {
        return new StringBuffer( "CREATE TABLE " ).append( sTableName )
                .append( "( id int(9) unsigned NOT NULL auto_increment, " )
                .append( "name varchar(255) NOT NULL default '', " )
                .append( "directory text, " )
                .append( "size bigint(18) unsigned NOT NULL default 0, " )
                .append( "lastmodified bigint(18) unsigned NOT NULL default 0, " )
                .append( "type char(1) NOT NULL default 'F', " )
                .append( "lastindexed bigint(18) unsigned NOT NULL default 0, " )
                .append( "node_id text NOT NULL, " )
                .append( "PRIMARY KEY  (id), " )
                .append( "FULLTEXT KEY `NameIndex` (`name`) ) TYPE=MyISAM;" ).toString();
    }

    /**
     * Builds and returns the SQL query string used to list all the tables that are
     * somehow like (or match) the given one in the current database. In MySQL, this
     * is done using "show tables like '&lt;sTableName&gt;'".
     *
     * @param sTableName  the name of the table to find the similar/equal ones
     * @return            the fully qualified MySQL string, as described above
     */
    public String getListAllTablesString( String sTableName )
    {
        return "show tables like '" + sTableName + "'";
    }

    /**
     * Builds and returns the SQL query string used to drop the given table.
     *
     * @param sTableName  the name of the table to be dropped
     * @return            the fully qualified SQL string
     */
    public String getDropTableString( String sTableName )
    {
        return "drop table " + sTableName;
    }

    /**
     * Builds and returns the SQL query string used to rename the given table.
     *
     * @param sOldTableName  the name of the table to be renamed
     * @param sNewTableName  the name of the table after the renaming
     * @return               the fully qualified SQL string
     */
    public String getRenameTableString( String sOldTableName, String sNewTableName )
    {
        return "alter table " + sOldTableName + " rename to " + sNewTableName;
    }

    /**
     * Builds and returns the SQL update string which updates the (unique) node with
     * the given id with the given update/reindexing time. This method is used to
     * update the parent node of a reindexed section which itself is not updated
     * during reindexing time.
     *
     * @param sTableName   the name of the table to be dropped
     * @param sNodeID      the id of the (unique) node to be updated with the given
     *      time
     * @param lUpdateTime  the time to set to the "lastindexed" field
     * @return             the fully qualified SQL string
     */
    public String getUpdateNodeIDString( String sTableName, String sNodeID, long lUpdateTime )
    {
        return new StringBuffer( "update " ).append( sTableName ).append( " set lastindexed = " )
                .append( lUpdateTime ).append( " where node_id = '" ).append( sNodeID ).append( "'" ).toString();
    }

    /**
     * Builds and returns the (sub)string of the final SQL string that is responsible
     * for result limitation on the database query's amount of results. For MySQL,
     * this looks like "LIMIT 0, 50", indicating that the results shall be listed
     * from the beginning (0), at most 50 results at a time.
     *
     * @return   the result-count-limitating SQL string (part), trimmed (without
     *      leading or ending spaces)
     */
    public String getLimitationString()
    {
        return new StringBuffer( "limit " ).append( this.getLowerLimit() ).append( ", " )
                .append( this.getUpperLimit() - this.getLowerLimit() ).toString();
    }

    /**
     * Builds and returns the SQL string which is needed to get the count of the
     * <i>child files</i> (no directories!) of the given node.
     *
     * @param sNodeID  the id of the (unique) node to get the children of
     * @return         the SQL string which runs the query that returns a single
     *      result column containing the amount of children of the given node
     */
    public String getChildCountString( String sNodeID )
    {
        return new StringBuffer( "select count(id) from " ).append( getFileListerConfig().getJDBCTableName() )
                .append( " where node_id like '" ).append( sNodeID ).append( "0%' and type = 'F'" ).toString();
    }

//-------------------------------- inner methods ------------------------------//
    /**
     * Appends the "where" keyword and the node id clause which makes sure that only
     * child entries of the given node id are included in the results.
     *
     * @param sb       the StringBuffer to append to
     * @param sNodeID  the node id of the directory to be searched
     */
    private void addNodeIDFilter( StringBuffer sb, String sNodeID )
    {
        //note: child nodes of sNodeID start with sNodeID itself, FOLLOWED BY a "0"
        sb.append( " where node_id like binary '" ).append( sNodeID ).append( "0%'" );
    }

    /**
     * Appends the type clause to the given StringBuffer, if only files or only directories
     * shall be listed.
     *
     * @param sb  the StringBuffer to append to
     */
    private void addTypeFilter( StringBuffer sb )
    {
        //add a type filter, if only one type shall be listed
        if ( !getFileListerConfig().isDirectoryListing() || !getFileListerConfig().isFileListing() )
            sb.append( " and type = '" ).append( ( getFileListerConfig().isDirectoryListing() ) ? "D'" : "F'" );
    }

    /**
     * Appends the "order by" clause, together with the various possible sort criteria,
     * each specified whether ascending or descending.
     *
     * @param sb  the StringBuffer to append to
     */
    private void addSorting( StringBuffer sb )
    {
        //create the sort order by adding the whole list of active criteria from
        //the FileListerConfig, each specified by asc or desc
        sb.append( " order by " );
        for ( int i = 0; i < getFileListerConfig().getSortOrders().size(); i++ )
        {
            sb.append( ( i != 0 ) ? ", " : "" );
            String sNextCriterium = ( String )getFileListerConfig().getSortOrders().get( i );
            if ( sNextCriterium.endsWith( "asc" ) )
                sb.append( sNextCriterium.substring( 0, sNextCriterium.length() - "asc".length() ) );
            else
                sb.append( sNextCriterium.substring( 0, sNextCriterium.length() - "desc".length() ) ).append( " desc" );
        }
    }

    /**
     * Appends the search string, together with the case-sensitivity, the exact match
     * and the regexp settings. The settings are retrieved from the given SearchParameters
     * instance.
     *
     * @param sb      the StringBuffer to append to
     * @param params  where to get the settings from
     */
    private void addSearchString( StringBuffer sb, SearchParameters params )
    {
        //if case-sensitive, use binary keyword
        String sBinary = ( params.isCaseSensitiveSearch() ) ? "binary " : "";

        //2 times 2 cases: regexp yes / no, exact match yes / no
        if ( !params.isUsingRegExp() )
        {
            if ( !params.isExactMatch() )
            {
                //no regular expressions, not an exact match: search for names containing the
                //search string, if *- and ?-free, else replace those two characters with
                //% and _, as used in MySQL; if the string itself contains an _, the search
                //string must be transformed into a regular expression...
                String sSearchString = params.getSearchString();
                boolean bNeedRegExp = ( sSearchString.indexOf( "_" ) >= 0 );
                String sLikeOrRegexp = ( bNeedRegExp ) ? "regexp " : "like ";
                boolean bHadModifiers = false;
                if ( sSearchString.indexOf( '*' ) >= 0 )
                {
                    sSearchString = sSearchString.replace( '*', ( bNeedRegExp ) ? '+' : '%' );
                    bHadModifiers = true;
                }
                if ( sSearchString.indexOf( '?' ) >= 0 )
                {
                    sSearchString = sSearchString.replace( '?', ( bNeedRegExp ) ? '.' : '_' );
                    bHadModifiers = true;
                }
                sb.append( " and name " ).append( sLikeOrRegexp ).append( sBinary )
                        .append( ( bNeedRegExp || bHadModifiers ) ? "'" : "'%" )
                        .append( sSearchString ).append( ( bNeedRegExp || bHadModifiers ) ? "'" : "%'" );
            }
            else
            {
                //no regular expressions, exact match: search for names equal to the search string
                sb.append( " and name like " ).append( sBinary ).append( "'" ).append(
                        params.getSearchString() ).append( "'" );
            }
        }
        else
        {
            //rem.: for the definition of MySQL regexp, see http://www.mysql.com/doc/R/e/Regexp.html
            //      for the usage, see http://www.mysql.com/doc/P/a/Pattern_matching.html
            if ( !params.isExactMatch() )
            {
                //regular expressions, not exact match: use the original search string (which
                //is treated as a regular expression)
                sb.append( " and name regexp " ).append( sBinary ).append( "'" ).append(
                        params.getSearchString() ).append( "'" );
            }
            else
            {
                //regular expressions, exact match: use the original search string (which
                //is treated as a regular expression) and emclose it with the regexp keyword '\b'
                sb.append( " and name regexp " ).append( sBinary ).append( "'^" ).append(
                        params.getSearchString() ).append( "$'" );
            }
        }
    }

    /**
     * Appends the minimum and the maximum size clauses, if defined by the user.
     *
     * @param sb      the StringBuffer to append to
     * @param params  where to get the settings from
     */
    private void addSizeConditions( StringBuffer sb, SearchParameters params )
    {
        if ( params.getMinimumSize() >= 0L )
            sb.append( " and size >= " + ( params.getMinimumSize() * params.getMinimumMultiplier() ) );
        if ( params.getMaximumSize() >= 0L )
            sb.append( " and size <= " + ( params.getMaximumSize() * params.getMaximumMultiplier() ) );
    }

    /**
     * Appends the starting and the ending date clauses, if defined by the user.
     *
     * @param sb      the StringBuffer to append to
     * @param params  where to get the settings from
     */
    private void addDateConditions( StringBuffer sb, SearchParameters params )
    {
        if ( params.getStartingDate() != null )
            sb.append( " and lastmodified >= " + params.getStartingDate().getTime() );
        if ( params.getEndingDate() != null )
            sb.append( " and lastmodified <= " + params.getEndingDate().getTime() );
    }
}
