package org.alltimeflashdreamer.filelister;

import java.io.Serializable;
import java.text.*;
import java.util.*;
import javax.servlet.http.HttpServletRequest;
import org.alltimeflashdreamer.util.StringUtils;

/**
 * <p>
 *
 * Title: SearchParameters</p> <p>
 *
 * Description: This class stores the settings of one FileLister search parameter
 * input/constellation. It is ready to be queried by SQLQueryBuilder or extending
 * classes to build the appropriate SQL query string(s).</p> <p>
 *
 * Organization: All Time Flash Dreamer</p>
 *
 * @author    Copyright (c) 2002-2004 by <a href="mailto:florian@steinsiepe.ch">Florian
 *      Steinsiepe</a>
 * @version   0.5.1
 */

public class SearchParameters implements Serializable
{
    /** place holder value for "not specified" or "not defined" (this is -1) */
    final static int NOTSPECIFIED = -1;

    private final static long[] MULTIPLIER_VALUES = {1L, 1024L, 1048576L, 1073741824L, 1099511627776L, 1125899906842624L};

    private static DateFormat df = DateFormat.getDateTimeInstance( DateFormat.SHORT, DateFormat.MEDIUM );
    private boolean bListDirectories = false;
    private boolean bListFiles = false;
    private boolean bUseRegExp;
    private boolean bCaseSensitiveSearch;
    private boolean bExactMatch;
    private String sSearchString;
    private long lMinimumSize = NOTSPECIFIED;
    private long lMinimumMultiplier = getDefaultMultiplier();
    private long lMaximumSize = NOTSPECIFIED;
    private long lMaximumMultiplier = getDefaultMultiplier();
    private Date dtStartingDate = null;
    private Date dtEndingDate = null;

    /**
     * Constructor reads the URL input parameters regarding the search settings that
     * need to be stored for each (consecutive) search.
     *
     * @param request               where to read the URL input from
     * @throws FileListerException  wrapper for each detected "bad input"
     */
    SearchParameters( HttpServletRequest request )
        throws FileListerException
    {
        //get the search string input (if empty, means "list all")
        setSearchString( request.getParameter( "searchwhat" ) );
        if ( StringUtils.isEmpty( getSearchString() ) )
            setSearchString( "" );

        //using regular expressions? case sensitive check? exakt match?
        setUsingRegExp( getBooleanInput( request.getParameter( "useregexp" ) ) );
        setCaseSensitiveSearch( getBooleanInput( request.getParameter( "casesensitive" ) ) );
        setExactMatch( getBooleanInput( request.getParameter( "exactmatch" ) ) );

        //get the selected file listing mode and check against abuse
        setFileListingMode( request.getParameter( "searchwhere" ) );

        //get the min and max size input, if any
        setSizeAndMultiplier( request.getParameter( "size_min" ), request.getParameter( "size_min_multiplier" ), true );
        setSizeAndMultiplier( request.getParameter( "size_max" ), request.getParameter( "size_max_multiplier" ), false );

        //get the sort order(s) input, if any
        setSortOrders( request );

        //get the starting and the ending dates, if any
        setStartingDate( getDateInput( request.getParameter( "startingdate" ), true ) );
        setEndingDate( getDateInput( request.getParameter( "endingdate" ), false ) );
    }

    /**
     * The empty constructor initializes the class with the application's default
     * values.
     */
    SearchParameters()
    {
        initializeDefaultValues();
    }

    /**
     * Tells whether the directories shall be listed.
     *
     * @return   list the directories?
     */
    public boolean isDirectoryListing()
    {
        return bListDirectories;
    }

    /**
     * Tells whether the files shall be listed.
     *
     * @return   list the files?
     */
    public boolean isFileListing()
    {
        return bListFiles;
    }

    /**
     * Returns whether the user selected to use regular expressions for the search
     * string or not.
     *
     * @return   user's using regexp?
     */
    public boolean isUsingRegExp()
    {
        return bUseRegExp;
    }

    /**
     * Returns whether the search string shall be handled in the case sensitive fashion
     * or not.
     *
     * @return   case sensitive compares?
     */
    public boolean isCaseSensitiveSearch()
    {
        return bCaseSensitiveSearch;
    }

    /**
     * Returns whether the search string shall be matched exactly or not.
     *
     * @return   case sensitive compares?
     */
    public boolean isExactMatch()
    {
        return bExactMatch;
    }

    /**
     * Returns the search string entered by the user for this specific search (cycle).
     *
     * @return   a String or "" (which represents "list all")
     */
    public String getSearchString()
    {
        return sSearchString;
    }

    /**
     * Returns whether all contents shall be listed (client entered the empty string
     * as the search string) or not.
     *
     * @return   whether to list all contents or filtered ones only
     */
    public boolean isListingAllEntries()
    {
        return ( StringUtils.isEmpty( sSearchString ) );
    }

    /**
     * Returns the currently set minimum value, as a String which is empty in case
     * no minimum size has yet been defined by the user.
     *
     * @return   the minimum size string or ""
     */
    public String getMinimumSizeOutput()
    {
        return ( ( lMinimumSize >= 0 ) ? "" + lMinimumSize : "" );
    }

    /**
     * Returns the currently set maximum value, as a String which is empty in case
     * no maximum size has yet been defined by the user.
     *
     * @return   the maximum size string or ""
     */
    public String getMaximumSizeOutput()
    {
        return ( ( lMaximumSize >= 0 ) ? "" + lMaximumSize : "" );
    }

    /**
     * Returns the currently set starting date value, as a String which is empty
     * in case no starting date size has yet been defined by the user.
     *
     * @return   the starting date string or ""
     */
    public String getStartingDateOutput()
    {
        return ( ( dtStartingDate == null ) ? "" : DateFormat.getDateInstance( DateFormat.SHORT ).format( dtStartingDate ) );
    }

    /**
     * Returns the currently set ending date value, as a String which is empty in
     * case no ending date size has yet been defined by the user.
     *
     * @return   the ending date string or ""
     */
    public String getEndingDateOutput()
    {
        return ( ( dtEndingDate == null ) ? "" : DateFormat.getDateInstance( DateFormat.SHORT ).format( dtEndingDate ) );
    }

    /**
     * Returns the whole list of option tags to be filled into the minimum size's
     * multiplier select box, having the correct entry selected.
     *
     * @return   the HTML string to be filled into the min size multiplier box
     */
    public String getMinimumMultiplierOutput()
    {
        return getMultiplierOptionList( lMinimumMultiplier );
    }

    /**
     * Returns the whole list of option tags to be filled into the maximum size's
     * multiplier select box, having the correct entry selected.
     *
     * @return   the HTML string to be filled into the max size multiplier box
     */
    public String getMaximumMultiplierOutput()
    {
        return getMultiplierOptionList( lMaximumMultiplier );
    }

    /**
     * Set the bListDirectories attribute.
     *
     * @param bListDirectories  list directories or not?
     */
    void setDirectoryListing( boolean bListDirectories )
    {
        this.bListDirectories = bListDirectories;
    }

    /**
     * Set the bListFiles attribute.
     *
     * @param bListFiles  list files or not?
     */
    void setFileListing( boolean bListFiles )
    {
        this.bListFiles = bListFiles;
    }

    /**
     * Sets the bUseRegExp attribute.
     *
     * @param bUseRegExp  use regexp or not?
     */
    void setUsingRegExp( boolean bUseRegExp )
    {
        this.bUseRegExp = bUseRegExp;
    }

    /**
     * Sets the bCaseSensitiveSearch attribute.
     *
     * @param bCaseSensitiveSearch  search case-sensitively or not?
     */
    void setCaseSensitiveSearch( boolean bCaseSensitiveSearch )
    {
        this.bCaseSensitiveSearch = bCaseSensitiveSearch;
    }

    /**
     * Sets the bExactMatch attribute.
     *
     * @param bExactMatch  search exact match or not?
     */
    void setExactMatch( boolean bExactMatch )
    {
        this.bExactMatch = bExactMatch;
    }

    /**
     * Sets the sSearchString attribute.
     *
     * @param sSearchString  search string of user?
     */
    void setSearchString( String sSearchString )
    {
        this.sSearchString = sSearchString;
    }

    /**
     * Returns the current minimum size value or -1, if not defined.
     *
     * @return   the minimum size or -1 (not defined)
     */
    long getMinimumSize()
    {
        return lMinimumSize;
    }

    /**
     * Sets the given minimum size value or -1, if lower than -1.
     *
     * @param lMinimumSize  the new minimum size
     */
    void setMinimumSize( long lMinimumSize )
    {
        this.lMinimumSize = ( lMinimumSize < NOTSPECIFIED ) ? NOTSPECIFIED : lMinimumSize;
    }

    /**
     * Returns the current maximum size value or -1, if not defined.
     *
     * @return   the maximum size or -1 (not defined)
     */
    long getMaximumSize()
    {
        return lMaximumSize;
    }

    /**
     * Sets the given maximum size value or -1, if lower than -1.
     *
     * @param lMaximumSize  the new maximum size
     */
    void setMaximumSize( long lMaximumSize )
    {
        this.lMaximumSize = ( lMaximumSize < NOTSPECIFIED ) ? NOTSPECIFIED : lMaximumSize;
    }

    /**
     * Returns the current starting date or null, if not defined.
     *
     * @return   the starting date or null
     */
    Date getStartingDate()
    {
        return dtStartingDate;
    }

    /**
     * Sets the given starting date.
     *
     * @param dtStartingDate  a new starting date or null, to unset it
     */
    void setStartingDate( Date dtStartingDate )
    {
        this.dtStartingDate = dtStartingDate;
    }

    /**
     * Returns the current ending date or null, if not defined.
     *
     * @return   the ending date or null
     */
    Date getEndingDate()
    {
        return dtEndingDate;
    }

    /**
     * Sets the given ending date.
     *
     * @param dtEndingDate  a new ending date or null, to unset it
     */
    void setEndingDate( Date dtEndingDate )
    {
        this.dtEndingDate = dtEndingDate;
    }

    /**
     * Returns the current minimum multiplier value.
     *
     * @return   the maximum multiplier
     */
    long getMinimumMultiplier()
    {
        return lMinimumMultiplier;
    }

    /**
     * Sets the current minimum multiplier value.
     *
     * @param lMinimumMultiplier  at least 1
     */
    void setMinimumMultiplier( long lMinimumMultiplier )
    {
        this.lMinimumMultiplier = ( lMinimumMultiplier >= 1 ) ? lMinimumMultiplier : getDefaultMultiplier();
    }

    /**
     * Returns the current maximum multiplier.
     *
     * @return   the maximum multiplier
     */
    long getMaximumMultiplier()
    {
        return lMaximumMultiplier;
    }

    /**
     * Sets the current maximum multiplier value.
     *
     * @param lMaximumMultiplier  at least 1
     */
    void setMaximumMultiplier( long lMaximumMultiplier )
    {
        this.lMaximumMultiplier = ( lMaximumMultiplier >= 1 ) ? lMaximumMultiplier : getDefaultMultiplier();
    }

    /** Initializes the class with default values (only). */
    void initializeDefaultValues()
    {
        setDirectoryListing( true );
        setFileListing( true );
        setCaseSensitiveSearch( false );
        setExactMatch( false );
        setUsingRegExp( false );
        setSearchString( "" );
        setMinimumSize( NOTSPECIFIED );
        setMinimumMultiplier( getDefaultMultiplier() );
        setMaximumSize( NOTSPECIFIED );
        setMaximumMultiplier( getDefaultMultiplier() );
        setStartingDate( null );
        setEndingDate( null );
    }

    /**
     * Returns true only if the given input equals "true", else false is returned.
     *
     * @param sURLInput  the input to parse
     * @return           a boolean value, as described above
     */
    boolean getBooleanInput( String sURLInput )
    {
        return ( !StringUtils.isEmpty( sURLInput ) && sURLInput.equals( "true" ) );
    }

    /**
     * Returns the default multiplier value (1024, indicating KB).
     *
     * @return   the default multiplier
     */
    static long getDefaultMultiplier()
    {
        return MULTIPLIER_VALUES[1];
    }

//--------------------------------------- private methods -------------------------------------//
    /**
     * Sets the "is file listing" and "is directory listing" booleans according to
     * the user input and checks against mismanipulations.
     *
     * @param sURLInput             the input to parse
     * @throws FileListerException  if the input is empty or not indicating either
     *      file or directory listing or both
     */
    private void setFileListingMode( String sURLInput )
        throws FileListerException
    {
        if ( StringUtils.isEmpty( sURLInput ) )
            throw new FileListerException( 0 );
        if ( sURLInput.equals( "both" ) )
        {
            setDirectoryListing( true );
            setFileListing( true );
        }
        else if ( sURLInput.equals( "files" ) )
        {
            setDirectoryListing( false );
            setFileListing( true );
        }
        else if ( sURLInput.equals( "directories" ) )
        {
            setDirectoryListing( true );
            setFileListing( false );
        }
        if ( !bListDirectories && !bListFiles )
            throw new FileListerException( 1 );
    }

    /**
     * Parses the given input into the minimum or a maximum size and multiplier values,
     * if defined by the user. Which one of both is defined by the given boolean.
     *
     * @param sInput                the input to parse into a size
     * @param sMultiplier           the input to parse into a multiplier
     * @param bIsMinimumSize        whether to read the minimum or the maximum value
     * @throws FileListerException  if the input isn't parseable/valid
     */
    private void setSizeAndMultiplier( String sInput, String sMultiplier, boolean bIsMinimumSize )
        throws FileListerException
    {
        //unset value if not contained in URL
        if ( StringUtils.isEmpty( sInput ) )
        {
            if ( bIsMinimumSize )
                setMinimumSize( NOTSPECIFIED );
            else
                setMaximumSize( NOTSPECIFIED );
            return;
        }

        //parse the multiplier
        try
        {
            long lMultiplier = new Long( sMultiplier ).longValue();
            boolean bFound = false;
            for ( int i = 0; i < MULTIPLIER_VALUES.length; i++ )
            {
                if ( MULTIPLIER_VALUES[i] == lMultiplier )
                {
                    bFound = true;
                    break;
                }
            }
            if ( !bFound )
                throw new FileListerException( 43 );
            if ( bIsMinimumSize )
                setMinimumMultiplier( lMultiplier );
            else
                setMaximumMultiplier( lMultiplier );
        } catch ( Exception ex )
        {
            if ( ex instanceof FileListerException )
                throw ( FileListerException )ex;

            throw new FileListerException( 43, ex );
        }

        //parse the user's size input
        try
        {
            long lInput = new Long( sInput ).longValue();
            if ( lInput < 0L )
                throw new FileListerException( 44 );
            if ( bIsMinimumSize )
                setMinimumSize( lInput );
            else
                setMaximumSize( lInput );
        } catch ( Exception ex )
        {
            if ( ex instanceof FileListerException )
                throw ( FileListerException )ex;

            throw new FileListerException( 44, ex );
        }
    }

    /**
     * Parses the given input into a starting or an ending date, as defined by the
     * given boolean.
     *
     * @param sURLInput             the input to parse
     * @param bIsStartingDate       starting or ending date?
     * @return                      the fully parsed/initialized date object
     * @throws FileListerException  if the input isn't parseable into a Date instance
     */
    private Date getDateInput( String sURLInput, boolean bIsStartingDate )
        throws FileListerException
    {
        //unset value if not contained in URL
        if ( StringUtils.isEmpty( sURLInput ) )
            return null;

        //parse the date input leniently; because this parsing demands a full date string
        //the time and timezone parameters must be added
        String sModInput = sURLInput + ( ( bIsStartingDate ) ? " 00:00:00.0" : " 23:59:59.999" );
        try
        {
            return DateFormat.getDateInstance( DateFormat.SHORT, Locale.US ).parse( sModInput );
        } catch ( Exception ex )
        {
            try
            {
                return DateFormat.getDateInstance( DateFormat.SHORT, Locale.GERMAN ).parse( sModInput );
            } catch ( Exception ex2 )
            {
                if ( ex2 instanceof ParseException )
                {
                    ParseException pe = ( ParseException )ex2;
                    pe = new ParseException( "Unparseable date: \"" + sURLInput + "\"", pe.getErrorOffset() );
                    throw new FileListerException( 45, pe );
                }

                throw new FileListerException( 45, ex2 );
            }
        }
    }

    /**
     * Returns a list of HTML option tags for the size multiplier select boxes. The
     * current multiplier setting/option is selected.
     *
     * @param lMultiplier  the multiplier to select within the list
     * @return             the full HTML option list string
     */
    private String getMultiplierOptionList( long lMultiplier )
    {
        StringBuffer sb = new StringBuffer();
        for ( int i = 0; i < StringUtils.MULTIPLIER_TEXTS.length; i++ )
        {
            sb.append( "<option value=\"" ).append( MULTIPLIER_VALUES[i] ).append( "\"" );
            if ( lMultiplier == MULTIPLIER_VALUES[i] )
                sb.append( " selected=\"on\"" );
            sb.append( ">" ).append( StringUtils.MULTIPLIER_TEXTS[i] ).append( "</option>" );
        }
        return sb.toString();
    }

    /**
     * Scans the given request for the entries starting with "sort" and collects
     * them into a list, if their value is not "none". The list is then passed to
     * FileListerConfig's setSortCriteria method.
     *
     * @param request  the request to scan for sort criteria
     */
    private void setSortOrders( HttpServletRequest request )
    {
        ArrayList arraySortOrders = new ArrayList();
        for ( int i = 0; i < FileListerConfig.SORTCRITERIA.length; i++ )
        {
            String sNextSortCriterium = request.getParameter( "sort" + ( i + 1 ) );
            if ( !StringUtils.isEmpty( sNextSortCriterium ) && !sNextSortCriterium.equals( "none" ) )
                arraySortOrders.add( sNextSortCriterium );
        }
        FileListerFactory.getFileListerConfig( request ).setSortCriteria( ( String[] )arraySortOrders.toArray( new String[]{} ) );
    }
}
